import dbm

from rltk.io.adapter import KeyValueAdapter
from rltk.io.serializer import Serializer, PickleSerializer
from rltk.record import Record


class DBMAdapter(KeyValueAdapter):
    def __init__(self, filename, dbm_class=dbm.ndbm, serializer: Serializer = None):
        """
        :dbm_class dbm, dbm.gnu, dbm.ndbm, dbm.dumb (same as dbm)
        """
        if not serializer:
            serializer = PickleSerializer()
        self._db = dbm_class.open(filename, 'c')
        self._serializer = serializer

        # not all dbm supports iterator, simulate an iterator by getting a key set
        # key will be in byte form once read out from dbm, needs to be converted to string
        self._ids = set([k.decode('utf-8') for k in self._db.keys()])

    def get(self, record_id):
        return self._serializer.loads(self._db.get(record_id))

    def set(self, record_id, record: Record):
        self._ids.add(record_id)
        self._db[record_id] = self._serializer.dumps(record)

    def __next__(self):
        for i in self._ids:
            yield self.get(i)

    def __del__(self):
        self._db.close()



import unicodedata


def check_for_none(*args):
    for arg in args:
        if arg is None:
            raise ValueError('Missing parameter')


def check_for_type(type, *args):
    for arg in args:
        if not isinstance(arg, type):
            raise TypeError('Wrong type of parameter')


def unicode_normalize(s):
    return unicodedata.normalize('NFKD', s)


def convert_list_to_set(s):
    if isinstance(s, list):
        s = set(s)
    return s



from rltk.record import Record


class KeyValueAdapter(object):
    def __init__(self):
        pass

    def __del__(self):
        pass

    def get(self, record_id) -> Record:
        raise NotImplementedError

    def set(self, record_id, record: Record):
        raise NotImplementedError

    def __iter__(self):
        return self.__next__()

    def __next__(self):
        """iterator is not required in adapter"""
        pass



from rltk.blocking import BlockGenerator


class CustomBlockGenerator(BlockGenerator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'custom_function' not in self._kwargs:
            raise ValueError('Missing argument: custom_function')
        self._custom_function = self._kwargs['custom_function']

    def _generate_blocks(self):
        for r1 in self._dataset1:
            for r2 in self._dataset2:
                if self._custom_function(r1, r2):
                    self._writer.write(r1.id, r2.id)



from rltk.io.adapter.key_value_adapter import KeyValueAdapter
from rltk.io.adapter.memory_adapter import MemoryAdapter
from rltk.io.adapter.dbm_adapter import DBMAdapter
from rltk.io.adapter.redis_adapter import RedisAdapter
from rltk.io.adapter.hbase_adapter import HBaseAdapter



from rltk.evaluation.trial import Trial


class Evaluation():
    def __init__(self):
        self.tp = 0
        self.tn = 0
        self.fp = 0
        self.fn = 0

    def add_trial(self, trial: Trial):
        self.tp, self.tn, self.fp, self.fn = self.__statistics_trial(trial)

    def precision(self) -> float:
        if self.tp + self.fp == 0:
            return 0
        return self.tp / (self.tp + self.fp)

    def recall(self) -> float:
        if self.tp + self.fn == 0:
            return 0
        return self.tp / (self.tp + self.fn)

    def f_measure(self) -> float:
        base = 2 * self.tp + self.fp + self.fn
        if base == 0:
            return 0
        return 2 * self.tp / base

    def false_positives(self) -> float:
        if (self.fp + self.tn) == 0:
            return 0
        return self.fp / (self.fp + self.tn)

    def true_positives(self) -> float:
        if (self.tp + self.fn) == 0:
            return 0
        return self.tp / (self.tp + self.fn)

    def false_negatives(self) -> float:
        if (self.tp + self.fn) == 0:
            return 0
        return self.fp / (self.tp + self.fn)

    def true_negatives(self) -> float:
        if (self.fp + self.tn) == 0:
            return 0
        return self.tn / (self.fp + self.tn)

    def __statistics_trial(self, trial: Trial):
        fp = 0
        fn = 0
        tp = 0
        tn = 0
        ground_truth = trial.get_ground_truth()

        for trial_result in trial.get_all_data():
            gt_val = ground_truth.is_positive(trial_result.record1.id, trial_result.record2.id)
            cal_val = trial_result.is_positive

            if cal_val and gt_val:
                tp += 1
            elif not cal_val and not gt_val:
                tn += 1
            elif cal_val and not gt_val:
                fp += 1
            elif not cal_val and gt_val:
                fn += 1

        return tp, tn, fp, fn



import os
import pytest
import redis

from rltk.record import Record
from rltk.io.adapter import *


class ConcreteRecord(Record):

    @property
    def id(self):
        return self.raw_object['id']

    @property
    def value(self):
        return self.raw_object['value']

record = ConcreteRecord(raw_object={'id': 'id1', 'value': 'value1'})


def test_memory_adapter():
    adapter = MemoryAdapter()
    adapter.set(record.id, record)
    assert adapter.get(record.id).id == record.id
    assert adapter.get(record.id).value == record.value
    for r in adapter:
        assert r.id == record.id
        break


def test_dbm_adapter():
    name = 'test_dbm_adapter'
    adapter = DBMAdapter(name)
    adapter.set(record.id, record)
    assert adapter.get(record.id).id == record.id
    assert adapter.get(record.id).value == record.value
    for r in adapter:
        assert r.id == record.id
        break
    if os.path.exists(name + '.db'):
        os.remove(name + '.db')


def test_redis_adapter():
    try:
        adapter = RedisAdapter('127.0.0.1', key_format='test_{record_id}')
        adapter.set(record.id, record)
        assert adapter.get(record.id).id == record.id
        assert adapter.get(record.id).value == record.value
        for r in adapter:
            assert r.id == record.id
            break

    except redis.exceptions.ConnectionError:
        return



import csv

from rltk.io.reader import Reader


class CSVReader(Reader):

    def __init__(self, file_handler, **kwargs):
        self._file_handler = Reader.get_file_handler(file_handler)
        self._csv_reader = csv.DictReader(self._file_handler, **kwargs)

    def __next__(self):
        for obj in self._csv_reader:
            yield {t[0]: t[1] for t in obj.items()}

    def __del__(self):
        try:
            self._file_handler.close()
        except:
            pass



import redis

from rltk.record import Record
from rltk.io.adapter import KeyValueAdapter
from rltk.io.serializer import Serializer, PickleSerializer


class RedisAdapter(KeyValueAdapter):
    def __init__(self, host, serializer: Serializer=None, key_format='{record_id}', **kwargs):
        if not serializer:
            serializer = PickleSerializer()
        self._redis = redis.Redis(host=host, **kwargs)
        self._serializer = serializer
        self._key_format = key_format
        try:
            self._get_key('test_id')
        except:
            raise ValueError('Invalid key_format.')

    def _get_key(self, record_id):
        return self._key_format.format(record_id=record_id)

    def __del__(self):
        pass

    def get(self, record_id) -> Record:
        return self._serializer.loads(self._redis.get(self._get_key(record_id)))

    def set(self, record_id, record: Record):
        return self._redis.set(self._get_key(record_id), self._serializer.dumps(record))

    def __iter__(self):
        return self.__next__()

    def __next__(self):
        # scan_iter() returns generator, keys() returns array
        for key in self._redis.scan_iter(self._get_key('*')):
            yield self._serializer.loads(self._redis.get(key))



from rltk.io.writer.block_writer import BlockWriter
from rltk.dataset import Dataset


class BlockGenerator(object):
    def __init__(self, dataset1 : Dataset, dataset2 : Dataset, writer: BlockWriter, **kwargs):
        self._writer = writer
        self._dataset1 = dataset1
        self._dataset2 = dataset2
        self._kwargs = kwargs

    def generate(self):

        self._generate_blocks()

        handler = self._writer.get_handler()
        self._writer.close()
        return handler

    def _generate_blocks(self):
        raise NotImplementedError



from rltk.io.writer import Writer


class BlockWriter(Writer):
    """
    index: sets
    black list is on index
    """

    def write(self, id1, id2):
        raise NotImplementedError

    def get_handler(self):
        raise NotImplementedError

    def flush(self):
        raise NotImplementedError

    def get_blacklist(self):
        return self._blacklist

    def close(self):
        self.flush()
        super().close()



from rltk.io.writer.writer import Writer
from rltk.io.writer.block_writer import BlockWriter
from rltk.io.writer.block_file_writer import BlockFileWriter
from rltk.io.writer.block_array_writer import BlockArrayWriter
from rltk.io.writer.ground_truth_writer import GroundTruthWriter



import io


class Reader(object):
    def __init__(self):
        pass

    def __iter__(self):
        return self.__next__()

    def __next__(self):
        """return raw content of one record"""
        raise NotImplementedError

    def __del__(self):
        pass

    @staticmethod
    def get_file_handler(f):
        if isinstance(f, io.IOBase):
            return f

        return open(f)



import pickle

from rltk.io.serializer import Serializer


class PickleSerializer(Serializer):

    def loads(self, string):
        return pickle.loads(string)

    def dumps(self, obj):
        return pickle.dumps(obj)



import json

from rltk.io.reader import GroundTruthReader
from rltk.io.writer import GroundTruthWriter


class GroundTruth(object):
    ID1 = 'id1'
    ID2 = 'id2'
    LABEL = 'label'

    def __init__(self):
        self.__ground_trurh_data = {}
        pass

    def add_positive(self, id1: str, id2: str):
        self.add_ground_truth(id1, id2, True)
        pass

    def add_negative(self, id1: str, id2: str):
        self.add_ground_truth(id1, id2, False)
        pass

    def add_ground_truth(self, id1: str, id2: str, value: bool):
        key = self.gen_key(id1, id2)
        self.__ground_trurh_data[key] = value

    def is_member(self, id1: str, id2: str) -> bool:
        key = self.gen_key(id1, id2)
        return key in self.__ground_trurh_data

    def is_positive(self, id1: str, id2: str) -> bool:
        key = self.gen_key(id1, id2)
        if self.__is_member(key):
            return self.__ground_trurh_data[key]
        else:
            raise Exception("not a member")

    def is_negative(self, id1: str, id2: str) -> bool:
        key = self.gen_key(id1, id2)
        if self.__is_member(key):
            return not self.__ground_trurh_data[key]
        else:
            raise Exception("not a member")

    def load(self, filename):
        # this will overwrite the current self.ground_trurh
        for obj in GroundTruthReader(filename):
            self.__ground_trurh_data[self.gen_key(obj[self.ID1], obj[self.ID2])] = obj[self.LABEL]

    def save(self, filename):
        w = GroundTruthWriter(filename)
        for k, v in self.__ground_trurh_data.items():
            ids = json.loads(k)
            w.write(ids[self.ID1], ids[self.ID2], v)
        w.close()

    def gen_key(self, id1: str, id2: str):
        key = json.dumps({self.ID1: id1, self.ID2: id2})
        return key

    def __is_member(self, key: str) -> bool:
        return key in self.__ground_trurh_data



import json
import heapq

from rltk.record import Record
from rltk.evaluation.ground_truth import GroundTruth


class Trial(object):
    class TrialResult:
        def __init__(self, record1: Record, record2: Record, is_positive: bool, confidence: float = None):
            self.record1 = record1
            self.record2 = record2
            self.is_positive = is_positive
            self.confidence = confidence

        def __cmp__(self, other):
            return self.confidence < other.confidence

        def __lt__(self, other):
            return self.confidence < other.confidence

    def __init__(self, groud_truth: GroundTruth, min_confidence: float = 0, top_k: int = 0):
        self.__ground_truth = groud_truth
        self.__min_confidence = min_confidence
        self.__top_k = top_k

        self.__data = []

    def add_result(self, record1: Record, record2: Record, is_positive: bool, confidence: float = None) -> None:
        if confidence >= self.__min_confidence and self.__ground_truth.is_member(record1.id, record2.id):
            if self.__top_k == 0 or len(self.__data) < self.__top_k:
                cur = self.TrialResult(record1, record2, is_positive, confidence)
                heapq.heappush(self.__data, cur)
            elif confidence > self.__data[0].confidence:
                heapq.heappop(self.__data)
                cur = self.TrialResult(record1, record2, is_positive, confidence)
                heapq.heappush(self.__data, cur)

    def add_positive(self, kwargs):
        """syntactic sugar"""
        self.add_result(is_positive=True, **kwargs)

    def add_negative(self, kwargs):
        """syntactic sugar"""
        self.add_result(is_positive=False, **kwargs)

    def get_all_data(self):
        return self.__data

    def get_ground_truth(self):
        return self.__ground_truth

    def __str__(self):
        res = json.dump(self.__data)
        return res

    def __repr__(self):
        pass



import collections
import math

import rltk.utils as utils


def tf_idf_similarity(bag1, bag2, df_corpus, doc_size, math_log=False):
    """
    Computes TF/IDF measure. This measure employs the notion of TF/IDF score commonly used in information retrieval (IR) to find documents that are relevant to keyword queries. The intuition underlying the TF/IDF measure is that two strings are similar if they share distinguishing terms.

    Args:
        bag1 (list): Bag 1.
        bag2 (list): Bag 2.
        df_corpus (dict): The pre calculated document frequency of corpus.
        doc_size (int): total documents used in corpus.
        math_log (bool, optional): Flag to indicate whether math.log() should be used in TF and IDF formulas. Defaults to False.

    Returns:
        float: TF/IDF cosine similarity.

    Examples:
        >>> rltk.tfidf(['a', 'b', 'a'], ['a', 'c'], {'a':3, 'b':1, 'c':1}, 3)
        0.17541160386140586
        >>> rltk.tfidf(['a', 'b', 'a'], ['a', 'c'], {'a':3, 'b':2, 'c':1}, 4, True)
        0.12977804138
        >>> rltk.tfidf(['a', 'b', 'a'], ['a'], {'a':3, 'b':1, 'c':1}, 3)
        0.5547001962252291
    """
    # http://www.tfidf.com/

    utils.check_for_none(bag1, bag2, df_corpus)
    utils.check_for_type(list, bag1, bag2)

    # term frequency for input strings
    t_x, t_y = collections.Counter(bag1), collections.Counter(bag2)
    tf_x = {k: float(v) / len(bag1) for k, v in t_x.items()}
    tf_y = {k: float(v) / len(bag2) for k, v in t_y.items()}

    # unique element
    total_unique_elements = set()
    total_unique_elements.update(bag1)
    total_unique_elements.update(bag2)

    idf_element, v_x, v_y, v_x_y, v_x_2, v_y_2 = 0.0, 0.0, 0.0, 0.0, 0.0, 0.0

    # tfidf calculation
    for element in total_unique_elements:
        if element not in df_corpus:
            continue
        idf_element = doc_size * 1.0 / df_corpus[element]

        v_x = 0 if element not in tf_x else (math.log(idf_element) * tf_x[element]) if math_log else (
            idf_element * tf_x[element])
        v_y = 0 if element not in tf_y else (math.log(idf_element) * tf_y[element]) if math_log else (
            idf_element * tf_y[element])
        v_x_y += v_x * v_y
        v_x_2 += v_x * v_x
        v_y_2 += v_y * v_y

    # cosine similarity
    return 0.0 if v_x_y == 0 else v_x_y / (math.sqrt(v_x_2) * math.sqrt(v_y_2))


def compute_tf(t, bag_len):
    """
    Args:
        t (dict): {term: count,...}
    """
    return {k: float(v) / bag_len for k, v in t.items()}


# # plus 1
# def compute_idf(df, doc_size, math_log=False):
#     return {k: float(doc_size) / v if math_log is False \
#             else math.log(float(doc_size + 1) / (v + 1)) \
#             for k, v in df.iteritems()}


def compute_idf(df, doc_size, math_log=False):
    return {k: doc_size * 1.0 / v if math_log is False \
        else math.log(doc_size * 1.0 / v) \
            for k, v in df.iteritems()}


def tf_idf_similarity_by_dict(tfidf_dict1, tfidf_dict2):
    """
    all terms of dict1 and dict2 should be in corpus

    tfidf_dict: {term: tfidf, ...}
    """
    v_x_y, v_x_2, v_y_2 = 0.0, 0.0, 0.0

    # intersection of dict1 and dict2
    # ignore the values that are not in both
    for t in tfidf_dict1.iterkeys():
        if t in tfidf_dict2:
            v_x_y = tfidf_dict1[t] * tfidf_dict2[t]

    for t, tfidf in tfidf_dict1.iteritems():
        v_x_2 += tfidf * tfidf
    for t, tfidf in tfidf_dict2.iteritems():
        v_y_2 += tfidf * tfidf

    # cosine similarity
    return 0.0 if v_x_y == 0 else v_x_y / (math.sqrt(v_x_2) * math.sqrt(v_y_2))



import json

from rltk.io.reader import Reader


class JsonLinesReader(Reader):

    def __init__(self, file_handler, ignore_blank_line=True):
        self._file_handler = Reader.get_file_handler(file_handler)
        self._ignore_blank_line = ignore_blank_line

    def __next__(self):
        for line in self._file_handler:
            if line.strip() == '':
                if self._ignore_blank_line:
                    continue
                else:
                    raise ValueError('Blank line detected')
            yield json.loads(line)

    def __del__(self):
        try:
            self._file_handler.close()
        except:
            pass



import os
import time
import tempfile

from rltk.blocking import BlockGenerator
from rltk.io.reader import BlockFileReader
from rltk.io.writer import BlockFileWriter


class InvertedIndexBlockGenerator(BlockGenerator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tokenizer = self._kwargs.get('tokenizer')
        self._tokenizer1 = self._kwargs.get('tokenizer1', self._tokenizer)
        self._tokenizer2 = self._kwargs.get('tokenizer2', self._tokenizer)
        self._buffer_size = self._kwargs.get('buffer_size') or 10000
        self._token_size = self._kwargs.get('token_size') or 1000
        self._temp_dir_path = self._kwargs.get('temp_dir_path', tempfile.gettempdir())

    def _generate_blocks(self):

        temp_file_prefix = os.path.join(self._temp_dir_path, str(time.time()))
        ds1_temp_filename = temp_file_prefix + '.ds1.temp'
        ds2_temp_filename = temp_file_prefix + '.ds2.temp'
        ds1_inverted_temp_filename = temp_file_prefix + '.ds1.inverted.temp'
        blacklist = set()

        # normal indices for ds1 (tokens: id), in order to complement blacklist
        ds1_temp_writer = BlockFileWriter(ds1_temp_filename, buffer_size=self._buffer_size,
                                          index_blacklist=blacklist, set_size=self._token_size)
        for r1 in self._dataset1:
            tokens = self._tokenizer1(r1)
            for t in tokens:
                ds1_temp_writer.write(t, r1.id)
        ds1_temp_writer.close()
        blacklist = ds1_temp_writer.get_blacklist()

        # inverted indices for ds1 (id: tokens)
        ds1_inverted_temp_writer = BlockFileWriter(ds1_inverted_temp_filename, buffer_size=self._buffer_size)
        for t, id1 in BlockFileReader(ds1_temp_filename):
            ds1_inverted_temp_writer.write(id1, t)
        ds1_inverted_temp_writer.close()

        # normal indices for ds2 (token: ids)
        ds2_temp_writer = BlockFileWriter(ds2_temp_filename, buffer_size=self._buffer_size,
                                          index_blacklist=blacklist, set_size=self._token_size)
        for r2 in self._dataset2:
            tokens = self._tokenizer2(r2)
            for t in tokens:
                ds2_temp_writer.write(t, r2.id)
        ds2_temp_writer.close()

        # construct blocks
        ds2_iter = iter(BlockFileReader(ds2_temp_filename))
        eof = False
        while not eof:
            # read a chunk of data from ds2 into memory
            ds2_indices = dict()
            count = self._buffer_size

            while count > 0:
                next_value = next(ds2_iter, None)
                if not next_value:
                    eof = True
                    break

                t, id2 = next_value
                if not t in ds2_indices:
                    ds2_indices[t] = set()
                ds2_indices[t] = ds2_indices[t] | set([id2])
                count -= 1

            # scan on-disk ds1 and in-memory ds2 chunk and make pairs
            ds1_inverted_temp_reader = BlockFileReader(ds1_inverted_temp_filename)
            for id1, t in ds1_inverted_temp_reader:
                if t in ds2_indices:
                    for id2 in ds2_indices[t]:
                        self._writer.write(id1, id2)

        os.remove(ds1_temp_filename)
        os.remove(ds2_temp_filename)
        os.remove(ds1_inverted_temp_filename)



from rltk.record import Record, cached_property, generate_record_property_cache, validate_record, remove_raw_object
from rltk.dataset import Dataset, get_record_pairs
from rltk.io import *
from rltk.similarity import *
from rltk.blocking import *
from rltk.tokenizer import *
from rltk.evaluation import *



import io


class Writer(object):
    def __init__(self):
        pass

    def write(self):
        raise NotImplementedError

    def __del__(self):
        pass

    def close(self):
        self.__del__()

    @staticmethod
    def get_file_handler(f):
        if isinstance(f, io.IOBase):
            return f

        return open(f, 'w')



from rltk.io.reader.block_reader import BlockReader
from rltk.io.writer.block_writer import BlockWriter


class BlockingHelper(object):
    def __init__(self, reader1 : BlockReader, reader2 : BlockReader):
        self._reader1 = reader1
        self._reader2 = reader2

    def union(self, writer: BlockWriter):

        for id1, id2 in self._reader1:
            writer.write(id1, id2)
        for id1, id2 in self._reader2:
            writer.write(id1, id2)
        writer.close()



from rltk.io.reader import Reader, BlockReader
from rltk.io.adapter import KeyValueAdapter, MemoryAdapter
from rltk.record import Record, generate_record_property_cache


class Dataset(object):
    """
    A set of records.
    
    Args:
        reader (Reader, optional): Input reader.
        record_class (type(Record), optional): Sub class of Record.
        adapter (KeyValueAdapter, optional): Specify where to store indexed data. Defaults to MemoryAdapter.
        size (int, optional): Maximum size of records will be in Dataset, defaults to None, can be used when debugging.
    Note:
        Set reader, record_class and adapter if new a Dataset needs to be generated.
        If Dataset is already generated and stored in a permanent adapter, only adapter needs to be provided.
    """
    def __init__(self, reader: Reader = None, record_class: type(Record) = None, adapter: KeyValueAdapter = None,
                 size: int = None):
        if not adapter:
            adapter = MemoryAdapter()
        self._adapter = adapter
        self._size = size

        # build index
        if reader and record_class:
            self._reader = reader
            self._record_class = record_class

            self._build_index()

    def _build_index(self):
        if not self._reader or not self._record_class:
            raise ValueError('Reader or Record class is not provided.')
        size = 0
        for raw_object in self._reader:
            record_instance = self._record_class(raw_object)
            generate_record_property_cache(record_instance)
            self._adapter.set(record_instance.id, record_instance)

            size += 1
            if self._size and size >= self._size:
                break

    def get_record(self, record_id):
        """
        Getter of a record.
        Args:
            record_id (str): Record id.
        Returns:
            Record: Record object.
        """
        return self._adapter.get(record_id)

    def __iter__(self):
        return self.__next__()

    def __next__(self):
        for r in self._adapter:
            yield r


def get_record_pairs(dataset1: Dataset, dataset2: Dataset, block_reader: BlockReader = None):
    """
    Generate pairs to compare.
    
    Args:
        dataset1 (Dataset): dataset 1
        dataset2 (Dataset): dataset 2
    """
    if not block_reader:
        for r1 in dataset1:
            for r2 in dataset2:
                yield r1, r2
    else:
        for id1, id2 in block_reader:
            yield dataset1.get_record(id1), dataset2.get_record(id2)



from distutils.core import setup
from setuptools import find_packages

with open('VERSION', 'r') as f:
    version = f.readline().strip()

with open('requirements.txt', 'r') as f:
    install_requires = list()
    for line in f:
        re = line.strip()
        if re:
            install_requires.append(re)

setup(
    name='rltk',
    version=version,
    packages=find_packages(),
    url='https://github.com/usc-isi-i2/rltk',
    license='MIT',
    author='USC/ISI',
    author_email='',
    description='Record Linkage ToolKit',
    package_data={'': ['LICENSE', 'VERSION', 'requirements.txt']},
    include_package_data=True,
    install_requires=install_requires
)



import csv

from rltk.io.writer import Writer


class GroundTruthWriter(Writer):
    def __init__(self, file_handler):
        self._file_handler = Writer.get_file_handler(file_handler)
        fieldnames = ['id1', 'id2', 'label']
        self._csv_writer = csv.DictWriter(self._file_handler, fieldnames=fieldnames)
        self._csv_writer.writeheader()

    def write(self, id1 : str, id2 : str, label : int):
        self._csv_writer.writerow({'id1': id1, 'id2': id2, 'label': label})

    def __del__(self):
        try:
            self._file_handler.close()
        except:
            pass



import json

from rltk.io.reader import Reader


class BlockFileReader(Reader):

    def __init__(self, file_handler):
        self._file_handler = Reader.get_file_handler(file_handler)

    def __next__(self):
        for obj in self._file_handler: # iterate json objects
            block = json.loads(obj)
            for id1, id2s in block.items(): # iterate on one json object
                for id2 in id2s:
                    yield id1, id2

    def __del__(self):
        try:
            self._file_handler.close()
        except:
            pass
